@@ -8,11 +8,20 @@ use base qw/Test::HTTP::Server::Simple HTTP::Server::Simple::CGI/;
 use Time::HiRes qw(sleep time);
 use Data::Dumper;
 use LWP::UserAgent;
+use Net::EmptyPort ();
 
 sub new {
     my ($class, $port) = @_;
+
+    if (!$port) {
+        $port = Net::EmptyPort::empty_port();
+    }
     
-    $port ||= 10_249; # randomish port
+    # Require a port parameter to be passed in.
+    # Any default here would mean the tests don't run properly in parallel.
+    if (!$port) {
+        die "Missing positional parameter 'port' required";
+    }
     
     return $class->SUPER::new($port);
 }
@@ -21,8 +30,8 @@ sub handle_request {
     my ( $self, $cgi ) = @_;
     my $params = $cgi->Vars;
 
-    # If we are on port 8081 then we are a proxy - we should forward the
-    # requests.
+    # If we should act as a proxy then the handle_request() behaviour is
+    # handled by act_as_proxy.
     return act_as_proxy(@_) if $self->{is_proxy};
 
     # We should act as a final destination server and so expect an absolute URL.
@@ -77,6 +86,15 @@ sub handle_request {
         print "Trickled for '$trickle_for'.\n";
     }
 
+    elsif ( exists $params->{cookie} ) {
+        print $cgi->header(
+            -nph => 1,
+            -cookie => $cgi->cookie(-name => "x", value => "test"),
+        );
+
+        print "Sent test cookie\n";
+    }
+
     elsif ( exists $params->{bad_header} ) {
         my $headers = $cgi->header( -nph => 1, );
 
